from abc import ABC
from typing import Generator, Type

from sequence.core.core import Sequence
from sequence.core.infinite_type import PropertyDefined
from sequence.core.utils.functions import is_prime


class PrimesOfSequence(PropertyDefined, ABC):
    """
    The class generates a sequence of prime numbers based on a specified property
    by filtering numbers from a base sequence.

    Attributes:
        base_sequence (Type[Sequence]): The base sequence used for filtering prime numbers.
    """

    def __init__(self, base_sequence: Type[Sequence]) -> None:
        super().__init__()
        self.base_sequence = base_sequence()

    def _as_generator(self) -> Generator:
        number = 1
        while True:
            if self.property(number=number) and number in self.base_sequence:
                yield number
            number += 1

    def property(self, number: int) -> bool:
        return is_prime(number=number)



from sequence.core.infinite_type import PropertyDefined
from sequence.core.utils.functions import is_prime
from sequence.sequences.integer.explicit import A000027
from sequence.sequences.integer.property_defined_generalised_sequences import PrimesOfSequence


class A000040(PrimesOfSequence):
    """ The prime numbers (https://oeis.org/A000040)."""
    sequence_name = 'prime numbers'

    def __init__(self) -> None:
        super().__init__(base_sequence=A000027)


PrimeNumbers = A000040
PositivePrimeNumbers = A000040


class A002808(PropertyDefined):
    """The composite numbers (https://oeis.org/A002808)."""
    sequence_name = 'composite numbers'

    def __init__(self) -> None:
        super().__init__()

    def property(self, number: int) -> bool:
        if number <= 3:
            return False
        return is_prime(number=number) is False


CompositeNumbers = A002808
PositiveCompositeNumbers = A002808



from typing import Any

from sequence.core.infinite_type import Explicit
from sequence.core.mixin import MonotonicIncreasingMixin
from sequence.sequences.integer.explicit_generalised_sequences import GeneralisedNexusNumbers, PolygonalNumbers


class A000027(Explicit):
    """The natural numbers (https://oeis.org/A000027)."""
    sequence_name = 'natural numbers'

    def __contains__(self, item: Any) -> bool:
        return True

    def formula(self, index: int) -> int:
        return index


PositiveIntegers = A000027
NaturalNumbers = A000027


class A000217(PolygonalNumbers):
    """Triangular numbers (https://oeis.org/A000217)."""
    sequence_name = 'triangular numbers'

    def __init__(self) -> None:
        super().__init__(number_of_sides=3)


TriangularNumbers = A000217


class A000290(Explicit):
    """Square numbers (https://oeis.org/A000290)."""
    sequence_name = 'square numbers'

    def __contains__(self, item: Any) -> bool:
        return False if item < 0 else int(item ** (1/2)) == item ** (1/2)

    def formula(self, index: int) -> int:
        return index ** 2


SquareNumbers = A000290


class A000326(PolygonalNumbers):
    """Pentagonal numbers (https://oeis.org/A000326)."""
    sequence_name = 'pentagonal numbers'

    def __init__(self) -> None:
        super().__init__(number_of_sides=5)


PentagonalNumbers = A000326


class A000384(PolygonalNumbers):
    """Hexagonal numbers (https://oeis.org/A000384)."""
    sequence_name = 'hexagonal numbers'

    def __init__(self) -> None:
        super().__init__(number_of_sides=6)


HexagonalNumbers = A000384


class A001045(MonotonicIncreasingMixin, Explicit):
    """Jacobsthal numbers (https://oeis.org/A001045)."""
    sequence_name = 'Jacobsthal numbers'

    def formula(self, index: int) -> int:
        return round(2 ** index / 3)


JacobsthalNumbers = A001045
JacobsthalSequence = A001045


class A003215(GeneralisedNexusNumbers):
    """Hex (or centered hexagonal) numbers (https://oeis.org/A003215)."""
    sequence_name = 'hex numbers'

    def __init__(self) -> None:
        super().__init__(dimension=2)

    def __contains__(self, item: Any) -> bool:
        if item <= 0:
            return False
        n = (3 + (12 * item - 3) ** (1 / 2)) / 6
        return n == int(n)


HexNumbers = A003215
CenteredHexagonalNumbers = A003215


class A005408(Explicit):
    """The odd numbers (https://oeis.org/A005408)."""
    sequence_name = 'odd numbers'

    def __contains__(self, item: Any) -> bool:
        return item % 2 == 1

    def formula(self, index: int) -> int:
        return 2 * index + 1


OddNumbers = A005408


class A014551(MonotonicIncreasingMixin, Explicit):
    """Jacobsthal-Lucas numbers (https://oeis.org/A014551)."""
    sequence_name = 'Jacobsthal-Lucas numbers'

    def __contains__(self, item: Any) -> bool:
        if item == 1:
            return True
        return super().__contains__(item=item)

    def formula(self, index: int) -> int:
        return 2**index + (-1)**index


JachobsthalLucasNumbers = A014551


class A033999(Explicit):
    """Sequence of powers of -1 (https://oeis.org/A033999)."""
    sequence_name = 'sequence of powers of -1'

    def __contains__(self, item: Any) -> bool:
        return item in {-1, 1}

    def formula(self, index: int) -> int:
        return (-1)**index



from abc import ABC, abstractmethod
from itertools import islice
from typing import Generator, List, Any, Optional, Union

from sequence.core.utils.exceptions import InfiniteSequenceError, NotPeriodicSequenceError
from sequence.core.utils.validation import validate_as_list_input


class Sequence(ABC):
    """Abstract base class for representing mathematical sequences."""
    sequence_name: str

    @abstractmethod
    def __contains__(self, item: Any) -> bool:
        raise NotImplementedError

    def __eq__(self, other: Any) -> bool:
        """
        The method checks if the representation of the sequence is the same.
        It doesn't check that two sequences are equal, but that two sequences are different instance of the same class.
        """
        return other.__str__() == self.__str__() and other.__dict__ == self.__dict__

    def __str__(self) -> str:
        return self.sequence_name

    def __iter__(self) -> Generator:
        return self._as_generator()

    def __getitem__(self, item: Any) -> Union[List, Any]:
        if isinstance(item, slice):
            if item.stop is None and self.is_finite is False:
                return islice(self._as_generator(), item.start, None)
            return self._as_list(start=item.start, stop=item.stop, step=item.step)
        return self._at(index=item)

    @abstractmethod
    def is_finite(self) -> bool:
        """Check if the sequence is finite."""
        raise NotImplementedError

    @property
    def is_periodic(self) -> bool:
        """Check if the sequence is periodic."""
        return False

    @property
    def period(self) -> int:
        """Get the period of the sequence. Raises NotPeriodicSequenceError if the sequence is not periodic."""
        raise NotPeriodicSequenceError

    @abstractmethod
    def _as_generator(self) -> Generator:
        """Return a generator for the sequence (internal use)."""
        raise NotImplementedError

    @abstractmethod
    def _as_list(self, stop: int, start: Optional[int] = None, step: Optional[int] = None) -> List[int]:
        """Return a list representation of the sequence within the specified range (internal use)."""
        raise NotImplementedError

    @abstractmethod
    def _at(self, index: int) -> Any:
        """Get the element at the specified index in the sequence (internal use)."""
        raise NotImplementedError


class FiniteType(Sequence, ABC):
    """Abstract base class for representing finite sequences."""

    def __init__(self, sequence: Optional[List[Any]] = None) -> None:
        super().__init__()
        self.sequence = sequence

    def __contains__(self, item: Any) -> bool:
        return item in self.sequence


class InfiniteType(Sequence, ABC):
    """Abstract base class for representing infinite sequences."""

    @property
    def is_finite(self) -> bool:
        return False

    def __len__(self) -> int:
        raise InfiniteSequenceError

    def _as_list(self, stop: int, start: Optional[int] = None, step: Optional[int] = None) -> List[int]:
        stop, start, step = validate_as_list_input(start=start, stop=stop, step=step)
        return list(islice(self, start, stop, step))

    def _at(self, index: int) -> int:
        return next(islice(self, index, index + 1))



from sequence.cli.cli import execute_command_line_interface

from sequence.sequences.integer.explicit import *
from sequence.sequences.integer.finite import *
from sequence.sequences.integer.periodic import *
from sequence.sequences.integer.property_defined import *
from sequence.sequences.integer.recursive import *


def main() -> None:
    execute_command_line_interface()


if __name__ == '__main__':
    main()



from sequence.core.finite_type import Periodic


class A087204(Periodic):
    """Sequence A087204 (https://oeis.org/A087204)."""
    sequence_name = 'sequence A087204'

    def __init__(self) -> None:
        super().__init__(_period=6)
        self.sequence = [2, 1, -1, -2, -1, 1]


class A128834(Periodic):
    """Sequence A128834 (https://oeis.org/A128834)."""
    sequence_name = 'sequence A128834'

    def __init__(self) -> None:
        super().__init__(_period=6)
        self.sequence = [0, 1, 1, 0, -1, -1]



from typing import Tuple

from sequence.core.utils.exceptions import NegativeNumberError


def validate_integer(integer: int) -> int:
    """The method validates that an object is an integer."""
    if isinstance(integer, int) is False:
        raise ValueError(f'Expected an int type, but got {type(integer).__name__} type!')
    return integer


def validate_positive_integer(integer: int) -> int:
    """The method validates that an object is a positive integer."""
    if validate_integer(integer=integer) < 0:
        raise NegativeNumberError(f'Expected a non-negative integer but got {integer}!')
    return integer


def validate_integer_tuple(tuple: Tuple[int, ...], length: int) -> None:
    """The method validates that an object is a tuple of integers."""
    if isinstance(tuple, Tuple) is False:
        raise ValueError(f'Expected a tuple type, but got {type(tuple).__name__} type!')
    if len(tuple) != length:
        raise ValueError(f'They are needed exactly {length} terms. Got {len(tuple)}')
    for number in tuple:
        validate_positive_integer(integer=number)


def validate_as_list_input(start: int, stop: int, step: int) -> Tuple[int, int, int]:
    """The method validates the input of the as_list method."""
    start = 0 if start is None else start
    step = 1 if step is None else step
    return stop, start, step




from sequence.cli.commands.list import execute_command_list
from sequence.cli.commands.sequence import execute_sequence_command
from sequence.cli.parser import CliParser


class CommandLineInterface:
    """A command-line interface for interacting with sequences."""

    @staticmethod
    def execute() -> None:
        """
        Parses command-line arguments using CliParser and executes either the print_list or execute_sequence_command
        based on the arguments.
        """
        args = CliParser().parse()

        if args.list:
            execute_command_list(args=args)
        else:
            execute_sequence_command(args=args)


def execute_command_line_interface() -> None:
    CommandLineInterface.execute()



import argparse
import importlib
import sys
from typing import Type

from sequence.core.core import Sequence


def execute_sequence_command(args: argparse.Namespace) -> None:
    """Executes a specific sequence command based on the provided arguments."""
    sequence = import_sequence_dynamically(sequence=args.sequence)
    if args.at is not None:
        execute_at_command(args=args, sequence=sequence)
    elif args.length is not None:
        execute_length_command(sequence=sequence)
    elif args.contains is not None:
        execute_contains_command(args=args, sequence=sequence)
    else:
        execute_as_list_command(args=args, sequence=sequence)


def import_sequence_dynamically(sequence: str) -> Type[Sequence]:
    """Dynamically import a sequence class by name."""
    try:
        sequence_module = importlib.import_module('sequence')
        sequence_class = getattr(sequence_module, f'{sequence}')
    except ImportError:
        sys.exit(f'Sequence {sequence} is not implemented yet!')
    else:
        return sequence_class()


def execute_as_list_command(args: argparse.Namespace, sequence: Type[Sequence]) -> None:
    """Execute the `as list` command, i.e., print a part of the sequence."""
    if args.start is None:
        args.start = 0
    if args.start < 0:
        sys.exit(f'invalid int value for --start: {args.start}. Expected a non-negative integer')
    if args.start > args.stop:
        sys.exit(f'invalid int value for --stop: {args.stop}. Expected a value bigger than --start')
    sys.exit(sequence[args.start:args.stop:args.step])


def execute_contains_command(args: argparse.Namespace, sequence: Type[Sequence]) -> None:
    sys.exit(f'{args.contains in sequence}')


def execute_length_command(sequence: Type[Sequence]) -> None:
    if sequence.is_finite:
        sys.exit(f'{len(sequence)}')
    else:
        sys.exit(f'{float("inf")}')


def execute_at_command(args: argparse.Namespace, sequence: Type[Sequence]) -> None:
    if args.at < 0:
        sys.exit(f'invalid int value for --start: {args.at}. Expected a non-negative integer!')
    else:
        sys.exit(f'{sequence[args.at]}')



import argparse
import re
import sys
from pathlib import Path

sequence_list_filepath = Path('sequence/SEQUENCES_LIST.md')


def execute_command_list(args: argparse.Namespace) -> None:
    """Prints a list of implemented sequences based on the provided argument."""
    with sequence_list_filepath.open() as file:
        content = file.read()

        if args.list == ['integer']:
            integer_table = extract_table(text=content, start_point="| OEIS    |")
            sys.exit('\n' + integer_table + '\n')

        if args.list == ['generalised']:
            # extract the part of text containing the table generalised sequences from the md
            generalised_sequences_pattern = re.compile(r'## Generalised sequences\n(.+?)(?=\n##|$)', re.DOTALL)
            match = re.search(generalised_sequences_pattern, content)
            text = match.group(1).strip()

            generalised_table = extract_table(text=text, start_point='| Sequence Name')
            sys.exit('\n' + generalised_table + '\n')


def extract_table(text: str, start_point: str) -> str:
    """Extract the table from a given text."""
    start_index = text.find(start_point)
    end_index = text.find("\n\n", text.find(start_point))
    return text[start_index:end_index].strip()




from abc import abstractmethod, ABC
from typing import Generator, Any, Tuple, List, Optional

from sequence.core.core import InfiniteType
from sequence.core.utils.validation import validate_positive_integer, validate_as_list_input


class Explicit(InfiniteType, ABC):
    """Abstract base class for representing explicit infinite sequences."""

    def __init__(self, start_index: int = 0) -> None:
        super().__init__()
        validate_positive_integer(integer=start_index)
        self.start_index = start_index

    def _as_generator(self) -> Generator:
        index = self.start_index
        while True:
            yield self.formula(index=index)
            index += 1

    @abstractmethod
    def formula(self, index: int) -> Any:
        """Abstract method to define the formula for generating elements in the sequence."""
        raise NotImplementedError

    def _as_list(self, stop: int, start: Optional[int] = None, step: Optional[int] = None) -> List[int]:
        stop, start, step = validate_as_list_input(start=start, stop=stop, step=step)
        return [self.formula(index=index) for index in range(start, stop, step)]

    def _at(self, index: int) -> int:
        validate_positive_integer(integer=index)
        return self.formula(index=index)


class Recursive(InfiniteType, ABC):
    """Abstract base class for representing recursive infinite sequences."""

    def __init__(self, start_terms: Optional[Tuple[Any, ...]] = None) -> None:
        super().__init__()
        self.start_terms = start_terms

    def _as_generator(self) -> Generator:
        terms = self.start_terms
        while True:
            yield terms[0]
            terms = self.formula(terms=terms)

    @abstractmethod
    def formula(self, terms: Tuple[Any, ...]) -> Tuple[Any, ...]:
        """Abstract method to define the recurrence relation of the sequence."""
        raise NotImplementedError


class PropertyDefined(InfiniteType, ABC):
    """Abstract base class for representing infinite sequences defined by a property."""

    def __contains__(self, item: Any) -> bool:
        return self.property(number=item)

    def _as_generator(self) -> Generator:
        number = 1
        while True:
            if self.property(number=number):
                yield number
            number += 1

    @abstractmethod
    def property(self, number: Any) -> bool:
        """Abstract method to define the property for elements in the sequence."""
        raise NotImplementedError



class NegativeNumberError(ValueError):
    """Error raise when trying to use a negative integer instead of a positive one."""


class InfiniteSequenceError(Exception):
    """Error raised when asking for finite sequence property to an infinite sequence."""


class NotPeriodicSequenceError(Exception):
    """Error raised when asking period to a non-periodic sequence."""



from typing import Any, List, ClassVar

from sequence.core.utils.functions import is_in_monotonic_increasing_generator


class MonotonicIncreasingMixin:
    """Mixin class for monotonic increasing sequences."""

    def __contains__(self, item: Any) -> bool:
        return is_in_monotonic_increasing_generator(generator=self, item=item)


class AlmostMonotonicIncreasingMixin:
    """Mixin class for almost monotonic increasing sequences."""
    offset: ClassVar[List[Any]]

    def __contains__(self, item: Any) -> bool:
        if item in self.offset:
            return True
        return is_in_monotonic_increasing_generator(generator=self[len(self.offset):], item=item)



from typing import Any

from sequence.core.infinite_type import Explicit
from sequence.core.mixin import MonotonicIncreasingMixin
from sequence.core.utils.validation import validate_positive_integer


class GeneralisedNexusNumbers(MonotonicIncreasingMixin, Explicit):
    """
    Class representing a sequence of Generalised Nexus Numbers (https://mathworld.wolfram.com/NexusNumber.html).

    Attributes:
        dimension (int): The dimension parameter for the Generalised Nexus Numbers.
    """
    sequence_name = 'generalised Nexus numbers'

    def __init__(self, dimension: int) -> None:
        super().__init__()
        self.dimension = validate_positive_integer(integer=dimension)

    def formula(self, index: int) -> int:
        return (index + 1) ** (self.dimension + 1) - index ** (self.dimension + 1)


class PolygonalNumbers(Explicit):
    """
        Represents a class for generating and working with polygonal numbers.

        Attributes:
        - number_of_sides (int): The number of sides for the polygonal numbers.
    """
    sequence_name = 'polygonal numbers'

    def __init__(self, number_of_sides: int) -> None:
        super().__init__()
        self.number_of_sides = validate_positive_integer(integer=number_of_sides)

    def __contains__(self, item: Any) -> bool:
        if item < 0:
            return False
        if item == 0:
            return True

        item = (
            (
                    8 * (self.number_of_sides - 2) * item + (self.number_of_sides - 4) ** 2
            ) ** (1/2)
            + (self.number_of_sides - 4)
        ) / (2 * (self.number_of_sides - 2))
        return item == int(item)

    def formula(self, index: int) -> int:
        return (self.number_of_sides - 2) * index * (index-1) // 2 + index



from typing import List

from sequence.core.finite_type import Finite
from sequence.core.utils.validation import validate_positive_integer


class A001228(Finite):
    """Orders of sporadic simple groups (https://oeis.org/A001228)."""
    sequence_name = 'orders of sporadic simple groups sequence'

    def __init__(self) -> None:
        super().__init__()
        self.sequence = [
            7920, 95040, 175560, 443520, 604800, 10200960, 44352000, 50232960, 244823040, 898128000, 4030387200,
            145926144000, 448345497600, 460815505920, 495766656000, 42305421312000, 64561751654400, 273030912000000,
            51765179004000000, 90745943887872000, 4089470473293004800, 4157776806543360000, 86775571046077562880,
            1255205709190661721292800, 4154781481226426191177580544000000,
            808017424794512875886459904961710757005754368000000000,
        ]


class A003173(Finite):
    """Heegner numbers: imaginary quadratic fields with unique factorization (https://oeis.org/A003173)."""
    sequence_name = 'Heegner numbers'

    def __init__(self) -> None:
        super().__init__()
        self.sequence = [1, 2, 3, 7, 11, 19, 43, 67, 163]


HeegnerNumbers = A003173


class CollatzSequence(Finite):
    """
    Collatz sequence: the sequence of numbers involved in the Collatz conjecture
    (https://en.wikipedia.org/wiki/Collatz_conjecture)
    """
    sequence_name = 'Collatz sequence'

    def __init__(self, start_value: int) -> None:
        super().__init__()
        start_value = validate_positive_integer(integer=start_value)
        self.sequence = self.generate_cycle(value=start_value)

    @staticmethod
    def generate_cycle(value: int) -> List[int]:
        cycle = [value]

        while value != 1:
            value = value // 2 if value % 2 == 0 else 3 * value + 1
            cycle.append(value)

        return cycle



from abc import ABC
from typing import List, Any, Generator, Optional

from sequence.core.core import FiniteType
from sequence.core.utils.exceptions import InfiniteSequenceError
from sequence.core.utils.validation import validate_as_list_input


class Finite(FiniteType, ABC):
    """Abstract base class for representing finite sequences."""

    def __init__(self) -> None:
        super().__init__()

    def __len__(self) -> int:
        return len(self.sequence)

    @property
    def is_finite(self) -> bool:
        return True

    def _as_generator(self) -> Generator:
        for element in self.sequence:
            yield element

    def _as_list(self, stop: int, start: int = 0, step: int = 1) -> List[int]:
        stop, start, step = validate_as_list_input(start=start, stop=stop, step=step)
        return self.sequence[start:stop:step]

    def _at(self, index: int) -> Any:
        return self.sequence[index]


class Periodic(FiniteType, ABC):
    """Abstract base class for representing periodic sequences."""

    def __init__(self, _period: int) -> None:
        super().__init__()
        self._period = _period

    def __len__(self) -> int:
        raise InfiniteSequenceError

    @property
    def is_finite(self) -> bool:
        return False

    @property
    def is_periodic(self) -> bool:
        return True

    @property
    def period(self) -> int:
        return self._period

    def _as_generator(self) -> Generator:
        index = 0
        while True:
            yield self.sequence[index]
            index = (index + 1) % self._period

    def _as_list(self, stop: int, start: Optional[int] = None, step: Optional[int] = None) -> List[int]:
        stop, start, step = validate_as_list_input(start=start, stop=stop, step=step)
        return [self.sequence[index % self._period] for index in range(start, stop, step)]

    def _at(self, index: int) -> Any:
        return self.sequence[index % self._period]




from math import isqrt
from typing import Generator, Any


def is_in_monotonic_increasing_generator(generator: Generator, item: Any) -> bool:
    """
    Check if the element 'item' is present in a monotonic increasing generator.

    Args:
        - generator (Generator)
        - item (Any): The item to be checked for presence in the generator
    """
    for element in generator:
        if element == item:
            return True
        if element > item:
            return False
    return False


def is_prime(number: int) -> bool:
    """
    Check if a given number is a prime number.

    Args:
        - number (int): The number to be checked.
    """
    if number <= 3:
        return number > 1
    if number % 2 == 0 or number % 3 == 0:
        return False

    return all(not (number % d == 0 or number % (d + 2) == 0) for d in range(5, isqrt(number) + 1, 6))



from typing import Any, Tuple, List, ClassVar
from sequence.core.mixin import AlmostMonotonicIncreasingMixin
from sequence.core.infinite_type import Recursive
from sequence.sequences.integer.recursive_generalised_sequences import (
    HighOrderFibonacciNumbers,
    LucasSequenceU,
    LucasSequenceV,
)


class A000032(LucasSequenceV):
    """Lucas numbers (https://oeis.org/A000032)."""
    sequence_name = 'Lucas numbers'

    def __init__(self) -> None:
        super().__init__(p=1, q=-1)

    def __contains__(self, item: Any) -> bool:
        if item == 1:
            return True
        return super().__contains__(item=item)


LucasNumbers = A000032


class A000045(LucasSequenceU):
    """Fibonacci numbers (https://oeis.org/A000045)."""
    sequence_name = 'Fibonacci numbers'

    def __init__(self) -> None:
        super().__init__(p=1, q=-1)


FibonacciNumbers = A000045
FibonacciSequence = A000045


class A000073(HighOrderFibonacciNumbers):
    """Tribonacci numbers (https://oeis.org/A000073)."""
    sequence_name = 'Tribonacci numbers'

    def __init__(self) -> None:
        super().__init__(order=3)


TribonacciNumbers = A000073


class A000078(HighOrderFibonacciNumbers):
    """Tetranacci numbers (https://oeis.org/A000078)."""
    sequence_name = 'Tetranacci numbers'

    def __init__(self) -> None:
        super().__init__(order=4)


TetranacciNumbers = A000078


class A000129(LucasSequenceU):
    """Pell numbers (https://oeis.org/A000129)."""
    sequence_name = 'Pell numbers'

    def __init__(self) -> None:
        super().__init__(p=2, q=-1)


PellNumbers = A000129
LambdaNumbers = A000129


class A000931(AlmostMonotonicIncreasingMixin, Recursive):
    """Padovan numbers (https://oeis.org/A000931)"""
    sequence_name = 'Padovan numbers'
    offset: ClassVar[List[int]] = [1, 0, 0, 1, 0, 1]

    def __init__(self) -> None:
        super().__init__(start_terms=(1, 0, 0))

    def formula(self, terms: Tuple[Any, ...]) -> Tuple[Any, ...]:
        return terms[1], terms[2], terms[1] + terms[0]


PadovanNumbers = A000931
PadovanSequence = A000931


class A001591(HighOrderFibonacciNumbers):
    """Pentanacci numbers (https://oeis.org/A001591)."""
    sequence_name = 'Pentanacci numbers'

    def __init__(self) -> None:
        super().__init__(order=5)


PentanacciNumbers = A001591


class A001592(HighOrderFibonacciNumbers):
    """Hexanacci numbers (https://oeis.org/A001592)."""
    sequence_name = 'Hexanacci numbers'

    def __init__(self) -> None:
        super().__init__(order=6)


HexanacciNumbers = A001591


class A002203(LucasSequenceV):
    """Companion Pell numbers (https://oeis.org/A002203)."""
    sequence_name = 'Companion Pell numbers'

    def __init__(self) -> None:
        super().__init__(p=2, q=-1)


CompanionPellNumbers = A002203
PellLucasNumbers = A002203


class A079262(HighOrderFibonacciNumbers):
    """Octanacci numbers (https://oeis.org/A079262)."""
    sequence_name = 'Octanacci numbers'

    def __init__(self) -> None:
        super().__init__(order=8)


OctanacciNumbers = A079262


class A104144(HighOrderFibonacciNumbers):
    """Enneanacci numbers (https://oeis.org/A104144)."""
    sequence_name = 'Enneanacci numbers'

    def __init__(self) -> None:
        super().__init__(order=9)


EnneanacciNumebrs = A104144


class A122189(HighOrderFibonacciNumbers):
    """Heptanacci numbers (https://oeis.org/A122189)."""
    sequence_name = 'Heptanacci numbers'

    def __init__(self) -> None:
        super().__init__(order=7)


HeptanacciNumbers = A122189


class A214733(LucasSequenceU):
    """Sequence A214733 (https://oeis.org/A214733)."""
    sequence_name = 'sequence A214733'

    def __init__(self) -> None:
        super().__init__(p=-1, q=3)

    def __contains__(self, item: Any) -> bool:
        if item == 0:
            return True
        for element in self._as_generator():
            if element == item:
                return True
            if abs(element) > abs(item * 1_000):
                return False
        return False



import argparse

from sequence.__version__ import __version__


class CliParser:
    """ A command-line interface parser"""

    def parse(self) -> argparse.Namespace:
        """Parses command-line arguments and returns the parsed arguments as a namespace."""
        parser = self.create_parser()
        parser = self.add_version_command(parser=parser)
        parser = self.add_groups_to_parser(parser=parser)
        parser = self.add_sequence_arguments(parser=parser)
        namespace = parser.parse_args()
        self.validate_parser(namespace=namespace, parser=parser)
        return namespace

    def create_parser(self) -> argparse.ArgumentParser:
        """Creates an ArgumentParser with program name, description, and epilog."""
        return argparse.ArgumentParser(
            prog='Sequentium',
            description="Sequentium is a user-friendly package that implements various well-known sequences, "
                        "providing a seamless and intuitive experience for the user",
            epilog='For help with a specific command, see: `sequence help <command>`',
        )

    def add_version_command(self, parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
        parser.add_argument(
            '-v',
            '--version',
            action='version',
            version=f'{parser.prog}: v{__version__}',
            help='Display the version information.',
        )
        return parser

    def add_groups_to_parser(self, parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
        """
        Adds mutually exclusive argument groups for version information, listing sequences,
        and sequence identification.
        """
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            '--list',
            nargs='+',
            choices=['integer', 'generalised'],
            help="List of implemented sequences.",
        )
        group.add_argument(
            'sequence',
            type=str,
            nargs='?',
            help='Specify the name or identifier of the sequence to operate on.',
        )
        return parser

    def add_sequence_arguments(self, parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
        """
        Adds arguments related to sequence operations such as retrieving terms, displaying length,
        defining start/stop points, specifying step size, and checking for a specific value.
        """
        sequence_group = parser.add_argument_group('sequence options')
        sequence_group.add_argument(
            '-a',
            '--at',
            type=int,
            default=None,
            help='Retrieve the term at the specified index in the sequence.',
        )
        sequence_group.add_argument(
            '-l',
            '--length',
            action='store_true',
            default=None,
            help='Display the length of the sequence.',
        )
        sequence_group.add_argument(
            "--start",
            type=int,
            default=None,
            help="Define the starting point of the sequence.",
        )
        sequence_group.add_argument(
            "--stop",
            type=int,
            default=None,
            help="End point of the sequence (excluded).",
        )
        sequence_group.add_argument(
            '--step',
            type=int,
            default=None,
            help='Step size for iterating through the sequence.',
        )
        sequence_group.add_argument(
            '-c',
            '--contains',
            type=int,
            default=None,
            help='Check if the sequence contains a specific value.',
        )
        return parser

    def validate_parser(
            self,
            namespace: argparse.Namespace,
            parser: argparse.ArgumentParser,
    ) -> argparse.Namespace:
        """Validate the provided namespace using the given parser."""
        if namespace.list is None and namespace.sequence is None:
            parser.error(message='No sequence or option provided. Please see --help')
        elif namespace.list is not None and namespace.sequence is not None:
            parser.error(message='argument --list: not allowed with argument sequence')
        elif namespace.list is not None and namespace.sequence is None:
            if len(namespace.list) > 1:
                parser.error(
                    message=f'argument --list: too many options provided. Expected 1, got {len(namespace.list)}'
                )
        else:
            namespace = self.validate_sequence_options(namespace=namespace, parser=parser)
        return namespace

    def validate_sequence_options(
            self,
            namespace: argparse.Namespace,
            parser: argparse.ArgumentParser,
    ) -> argparse.Namespace:
        """Validate the selected options for a sequence in the provided namespace using the given parser."""
        selected_options = {key for key, value in vars(namespace).items() if value is not None}
        selected_options.remove('sequence')

        if len(selected_options) == 0:
            parser.error(message=f'No options chosen for sequence {namespace.sequence.__str__()}')
        elif len(selected_options) > 1:  # noqa: SIM102
            if 'at' in selected_options or 'c' in selected_options or 'length' in selected_options:
                parser.error(message=f'options {list(selected_options)}: not allowed together')

        return namespace



from pathlib import Path
from setuptools import setup, find_namespace_packages

from sequence.__version__ import __version__

long_description = Path('README.md').read_text()

setup(
  name='sequentium',
  version=f'{__version__}',
  license='MIT',
  author='Vasco Schiavo',
  author_email='vasco.schiavo@protonmail.com',
  url='https://github.com/VascoSch92/sequentium',
  description="Sequentium is a user-friendly package that implements various well-known sequences, providing a "
              "seamless and intuitive experience for the user",
  long_description=long_description,
  long_description_content_type='text/markdown',
  packages=find_namespace_packages(include=['sequence*'], exclude=['tests']),
  entry_points={
        'console_scripts': ['sequence=sequence:main']
    },
  # download_url='https://github.com/user/reponame/archive/v_01.tar.gz',
  keywords=['math', 'mathematics', 'sequence', 'fibonacci'],
  classifiers=[
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Education',
    'Intended Audience :: End Users/Desktop',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3.12',
    'Topic :: Education',
    'Topic :: Scientific/Engineering :: Mathematics',
  ],
  python_requires='>=3.9',
)




from typing import Tuple, Any

from sequence.core.infinite_type import Recursive
from sequence.core.mixin import MonotonicIncreasingMixin
from sequence.core.utils.validation import validate_integer, validate_positive_integer


class HighOrderFibonacciNumbers(MonotonicIncreasingMixin, Recursive):
    """
        Class for generating high order Fibonacci numbers (https://mathworld.wolfram.com/Fibonaccin-StepNumber.html).

        Attributes:
            order (int): The order of the Fibonacci sequence.
    """
    sequence_name = 'High order Fibonacci numbers'

    def __init__(self, order: int) -> None:
        self.order = validate_positive_integer(integer=order)
        start_terms = self._compute_start_terms()

        super().__init__(start_terms=start_terms)

    def _compute_start_terms(self) -> Tuple[Any, ...]:
        """
        The method gives a tuple representing the start terms of the Fibonacci sequence.
        The start terms are computed as following:
            If the Fibonacci sequence has order n, then the tuple has 0 in the n-1 first position and 1 in the last.
        """
        return (0,) * (self.order - 1) + (1,)

    def formula(self, terms: Tuple[Any, ...]) -> Tuple[Any, ...]:
        return *[terms[i] for i in range(1, self.order)], sum(terms)


class LucasSequenceU(MonotonicIncreasingMixin, Recursive):
    """
    The class generates the Lucas sequence U_n (https://en.wikipedia.org/wiki/Lucas_sequence).
    The sequence is defined by the recurrence relation: U_{n+2} = p * U_{n+1} - q * U_n,
    with initial terms (0, 1).

    Attributes:
        p (int): Coefficient for the U_{n+1} term.
        q (int): Coefficient for the U_n term.
    """
    sequence_name = 'Lucas sequence U_n'

    def __init__(self, p: int, q: int) -> None:
        super().__init__(start_terms=(0, 1))
        self.p = validate_integer(integer=p)
        self.q = validate_integer(integer=q)

    def formula(self, terms: Tuple[int, int]) -> Tuple[int, int]:
        return terms[1], self.p * terms[1] - self.q * terms[0]


class LucasSequenceV(LucasSequenceU):
    """
    The class extends LucasSequenceU and generates the Lucas sequence V_n
    (https://en.wikipedia.org/wiki/Lucas_sequence), a variant with different initial terms.
    The sequence is defined by the recurrence relation: V_{n+2} = p * V_{n+1} - q * V_n,
    with initial terms (2, p).

    Attributes:
        p (int): Coefficient for the V_{n+1} term.
        q (int): Coefficient for the V_n term.
    """
    sequence_name = 'Lucas sequence V_n'

    def __init__(self, p: int, q:  int) -> None:
        super().__init__(p=p, q=q)

        self.start_terms = (2, self.p)
